'''Wraps the user-defined interaction into a class that stores several relevant
parameters of the problem.
'''

from typing import Callable
import numpy as np
from .constants import HBARC, ALPHA
from .spin_orbit import SpinOrbitTerm

class Interaction:
    '''Defines a local, (possibly) complex, affine, fixed-energy interaction.
    '''
    def __init__(self,
        coordinate_space_potential: Callable[[float, np.array], float], # V(r, theta)
        n_theta: int, # How many parameters does the interaction have?
        mu: float, # reduced mass (MeV)
        energy: float, # E_{c.m.}
        ell: int,
        Z_1: int = 0, # atomic number of particle 1
        Z_2: int = 0, # atomic number of particle 2
        R_C: float = 0.0, # Coulomb "cutoff"
        is_complex: bool = False,
        spin_orbit_term: SpinOrbitTerm = None,
    ):
        r'''Creates a local, (possibly) complex, affine, fixed-energy interaction.

        Parameters:
            coordinate_space_potential (Callable[[float,ndarray],float]): V(r, theta)
            n_theta (int): number of parameters
            mu (float): reduced mass
            energy (float): center-of-mass, scattering energy
            ell (int): angular momentum
            Z_1 (int): charge of particle 1
            Z_2 (int): charge of particle 2
            R_C (float): Coulomb "cutoff" radius
            is_complex (bool): Is the interaction complex?
            spin_orbit_term (SpinOrbitTerm): See [Spin-Orbit section](#spin-orbit).
        
        Returns:
            instance (Interaction): instance of `Interaction`

        Attributes:
            v_r (Callable[[float,ndarray],float]): coordinate-space potential; $V(r, \alpha)$
            n_theta (int): number of parameters
            mu (float): reduced mass
            ell (int): angular momentum
            k_c (float): Coulomb momentum; $k\eta$
            is_complex (bool): Is this a complex potential?
            spin_orbit_term (SpinOrbitTerm): See [Spin-Orbit section](#spin-orbit)

        '''
        self.v_r = coordinate_space_potential
        self.n_theta = n_theta
        self.mu = mu / HBARC # Go ahead and convert to 1/fm
        self.ell = ell
        self.k_c = ALPHA * Z_1*Z_2 * self.mu
        self.R_C = R_C
        # self.eta = ALPHA * Z_1 * Z_2 * self.mu / self.k
        self.is_complex = is_complex
        self.spin_orbit_term = spin_orbit_term

        if spin_orbit_term is None:
            self.include_spin_orbit = False
        else:
            self.include_spin_orbit = True

        if energy:
            # If the energy is specified (not None as it is when subclass
            # EnergizedInteraction instantiates), set up associated attributes.
            self.energy = energy
            self.k = np.sqrt(2 * self.mu * self.energy/HBARC)
            self.sommerfeld = self.k_c / self.k
        else:
            # If the energy is not specified, these will be set up when the
            # methods are called.
            self.energy = None
            self.k = None
            self.sommerfeld = None


    def tilde(self,
        s: float,
        alpha: np.array
    ):
        r'''Scaled potential, $\tilde{U}(s, \alpha, E)$.

        * Does not include the Coulomb term.
        * $E = E_{c.m.}$; `[E] = MeV = [v_r]`

        Parameters:
            s (float): mesh point; $s = pr/\hbar$
            alpha (ndarray): the varied parameters
        
        Returns:
            u_tilde (float | complex): value of scaled interaction

        '''
        vr = self.v_r(s/self.k, alpha)
        vr += self.spin_orbit_term.spin_orbit_potential(s/self.k, alpha) if self.include_spin_orbit else 0
        return  1.0/self.energy * vr


    def basis_functions(self,
        rho_mesh: np.array
    ):
        r'''In general, we approximate the potential as

        $\hat{U} = \sum_{j} \beta_j(\alpha) u_j$
        
        For affine interactions (like those defined in this class) the basis
        functions (or "pillars), $u_j$, are just the "naked" parts of the
        potential. As seen below, it is assumed that the $\beta_j(\alpha)$
        coefficients are just the affine parameters, $\alpha$, themselves.

        Parameters:
            rho_mesh (ndarray): discrete $\rho$ values at which the potential is
                going to be evaluated
        
        Returns:
            value (ndarray): values of the scaled potential at provided $\rho$ points
        '''
        return np.array([
            self.tilde(rho_mesh, row) for row in np.eye(self.n_theta)
        ]).T
    

    def coefficients(self,
        alpha: np.array # interaction parameters
    ):
        r'''As noted in the `basis_functions` documentation, the coefficients
        for affine interactions are simply the parameter values. The inverse of
        the momentum is also returned to support energy emulation.

        Parameters:
            alpha (ndarray): parameter point
        
        Returns:
            result (tuple): inverse momentum and coefficients

        '''
        return 1/self.k, alpha
    

    def eta(self,
        alpha: np.array
    ):
        r'''Sommerfeld parameter. Implemented as a function to support energy
        emulation (where the energy could be a part of the parameter vector,
        `alpha`).
        
        Parameters:
            alpha (ndarray): parameter vector
        
        Returns:
            eta (float): Sommerfeld parametere
        '''
        return self.sommerfeld
    

    def momentum(self, alpha: np.array):
        r'''Momentum. Implemented as a function to support energy emulation
        (where the energy/momentum could be a part of the parameter vector,
        `alpha`).
        
        Parameters:
            alpha (ndarray): parameter vector
        
        Returns:
            k (float): center-of-mass, scattering momentum
        '''
        return self.k


class InteractionSpace:
    def __init__(self,
        coordinate_space_potential: Callable[[float, np.array], float], # V(r, theta)
        n_theta: int, # How many parameters does the interaction have?
        mu: float, # reduced mass (MeV)
        energy: float, # E_{c.m.}
        l_max: int,
        Z_1: int = 0, # atomic number of particle 1
        Z_2: int = 0, # atomic number of particle 2
        R_C: float = 0.0,
        is_complex: bool = False,
        spin_orbit_potential: Callable[[float, np.array, float], float] = None #V_{SO}(r, theta, l•s)
    ):
        r'''Generates a list of $\ell$-specific interactions.
        
        Parameters:
            coordinate_space_potential (Callable[[float,ndarray],float]): V(r, theta)
            n_theta (int): number of parameters
            mu (float): reduced mass
            energy (float): center-of-mass, scattering energy
            l_max (int): maximum angular momentum
            Z_1 (int): charge of particle 1
            Z_2 (int): charge of particle 2
            R_C (float): Coulomb "cutoff" radius
            is_complex (bool): Is the interaction complex?
            spin_orbit_potential (Callable[[float,ndarray,float],float]): coordinate-space, spin-orbit potential; $V_{\rm SO}(s, \alpha, 2\ell\cdot s$)
        
        Returns:
            instance (InteractionSpace): instance of InteractionSpace
        
        Attributes:
            interaction (list): list of `Interaction`s
        '''
        self.interactions = []
        if spin_orbit_potential is None:
            for l in range(l_max+1):
                self.interactions.append(
                    [Interaction(coordinate_space_potential, n_theta, mu,
                        energy, l, Z_1=Z_1, Z_2=Z_2, R_C=R_C,
                        is_complex=is_complex)]
                )
        else:
            for l in range(l_max+1):
                self.interactions.append(
                    [Interaction(coordinate_space_potential, n_theta, mu,
                        energy, l, Z_1=Z_1, Z_2=Z_2, R_C=R_C,
                        is_complex=is_complex,
                        spin_orbit_term=SpinOrbitTerm(spin_orbit_potential,
                        lds)) for lds in couplings(l)]
                )


def couplings(l):
    r'''For a spin-1/2 nucleon scattering off a spin-0 nucleus, there are
    maximally 2 different total angular momentum couplings: l+1/2 and l-1/2.
    
    Parameters:
        l (int): angular momentum
    
    Returns:
        couplings (list): total angular momentum possibilities
    '''
    if l == 0:
        return [l]
    else:
        return [l, -(l+1)]




r'''A ReducedBasisEmulator builds the reduced basis for a given interaction and
partial wave.

It instantiates a Basis and precomputes the integrals it can with the basis
states and operators. When the user asks for an emulated phase shift, the
emulator computes the $\hat{\phi}$ coefficients.
'''
import pickle
import numpy as np

from .interaction import Interaction
from .schroedinger import SchroedingerEquation
from .basis import RelativeBasis, Basis, CustomBasis
from .constants import DEFAULT_RHO_MESH, HBARC
from .free_solutions import phase_shift, H_minus, H_plus, H_minus_prime, H_plus_prime
from .utility import finite_difference_first_derivative, finite_difference_second_derivative, regular_inverse_s

class ReducedBasisEmulator:
    r'''A ReducedBasisEmulator (RBE) uses the specified `interaction` and
    `theta_train` to generate solutions to the Schrödinger equation at a
    specific energy (`energy`) and partial wave (`l`).

    Using the Galerkin projection method, a linear combination of those
    solutions (or a PCA basis of them) is found at some arbitrary point in
    parameter space, `theta`.

    '''
    @classmethod
    def load(obj, filename):
        r'''Loads a previously trained emulator.

        Parameters:
            filename (string): name of file
        
        Returns:
            emulator (ReducedBasisEmulator): previously trainined `ReducedBasisEmulator`
        
        '''
        with open(filename, 'rb') as f:
            rbe = pickle.load(f)
        return rbe


    @classmethod
    def from_train(cls,
        interaction: Interaction,
        theta_train: np.array, # training points in parameter space
        n_basis: int = 4, # How many basis vectors?
        use_svd: bool = True, # Use principal components as basis vectors?
        s_mesh: np.array = DEFAULT_RHO_MESH, # s = rho = kr; solutions are phi(s)
        s_0: float = 6*np.pi, # phase shift is "extracted" at s_0
        hf_tols: list = None, # 2 numbers: high-fidelity solver tolerances, relative and absolute
    ):
        r'''Trains a reduced-basis emulator based on the provided interaction and training space.
        
        Parameters:
            interaction (Interaction): local interaction
            theta_train (ndarray): training points in parameter space; shape = (n_points, n_parameters)
            n_basis (int): number of basis vectors for $\hat{\phi}$ expansion
            use_svd (bool): Use principal components of training wave functions?
            s_mesh (ndarray): $s$ (or $\rho$) grid on which wave functions are evaluated
            s_0 (float): $s$ point where the phase shift is extracted
            hf_tols (list): 2-element list passed to `SchroedingerEquation`;
                [relative tolerance, absolute tolerance]
        
        Returns:
            rbe (ReducedBasisEmulator): $\ell$-specific, reduced basis emulator

        '''
        basis = RelativeBasis(
            SchroedingerEquation(interaction, hifi_tolerances=hf_tols),
            theta_train, s_mesh, n_basis, interaction.ell, use_svd
        )
        return cls(interaction, basis, s_0=s_0)


    def __init__(self,
        interaction: Interaction,
        basis: Basis,
        s_0: float = 6*np.pi # phase shift is "extracted" at s_0
    ):
        r'''Trains a reduced-basis emulator based on the provided interaction and basis.
        
        Parameters:
            interaction (Interaction): local interaction
            basis (Basis): see [Basis documentation](basis.md)
            s_0 (float): $s$ point where the phase shift is extracted
        
        Returns:
            rbe (ReducedBasisEmulator): $\ell$-specific, reduced basis emulator

        '''
        self.interaction = interaction
        self.basis = basis
        self.l = self.basis.l

        if isinstance(self.basis, CustomBasis):
            self.basis.solver = basis.solver

        self.s_mesh = np.copy(basis.rho_mesh)

        # Index of the point in the s mesh that is closest to s_0.
        self.i_0 = np.argmin(np.abs(self.s_mesh - s_0))
        # We want to choose a point at which the solution has already been
        # calculated so we can avoid interpolation.
        self.s_0 = self.s_mesh[self.i_0]
        if self.interaction.k_c == 0:
            # No Coulomb, so we can precompute with eta = 0.
            self.Hm = H_minus(self.s_0, self.l, 0)
            self.Hp = H_plus(self.s_0, self.l, 0)
            self.Hmp = H_minus_prime(self.s_0, self.l, 0)
            self.Hpp = H_plus_prime(self.s_0, self.l, 0)
        else:
            # There is Coulomb, but we haven't (yet) worked out how to emulate
            # across energies, so we can precompute H+ and H- stuff.
            eta = self.interaction.eta(None)
            self.Hm = H_minus(self.s_0, self.l, eta)
            self.Hp = H_plus(self.s_0, self.l, eta)
            self.Hmp = H_minus_prime(self.s_0, self.l, eta)
            self.Hpp = H_plus_prime(self.s_0, self.l, eta)

        # \tilde{U}_{bare} takes advantage of the linear dependence of \tilde{U}
        # on the parameters. The first column is multiplied by args[0]. The
        # second by args[1]. And so on. The "total" potential is the sum across
        # columns.
        self.utilde_basis_functions = self.interaction.basis_functions(self.s_mesh)

        # Precompute what we can for < psi | F(hat{phi}) >.
        d2_operator = finite_difference_second_derivative(self.s_mesh)
        phi_basis = self.basis.vectors
        ang_mom = self.l*(self.l+1) / self.s_mesh**2

        # This is not going to work when we implement energy emulation with Coulomb.
        S_C = self.interaction.R_C * self.interaction.k
        k_c = 2*self.interaction.k_c * regular_inverse_s(self.s_mesh,S_C )

        self.d2 = -d2_operator @ phi_basis
        self.A_1 = phi_basis.T @ self.d2
        self.A_2 = np.array([
            phi_basis.T @ (row[:, np.newaxis] * phi_basis) for row in self.utilde_basis_functions.T
        ])
        self.A_3 = np.einsum('ij,j,jk',
                             phi_basis.T,
                             ang_mom - 1,
                             phi_basis)
        self.A_13 = self.A_1 + self.A_3
        self.A_3_coulomb = np.einsum('ij,j,jk', phi_basis.T, k_c, phi_basis)

        # Precompute what we can for the inhomogeneous term ( -< psi | F(phi_0) > ).
        d2_phi_0 = d2_operator @ self.basis.phi_0
        self.b_1 = phi_basis.T @ d2_phi_0
        self.b_2 = np.array([
            phi_basis.T @ (-row * self.basis.phi_0) for row in self.utilde_basis_functions.T
        ])
        self.b_3 = phi_basis.T @ ((1 - ang_mom) * self.basis.phi_0)
        self.b_13 = self.b_1 + self.b_3
        self.b_3_coulomb = -phi_basis.T @ (k_c * self.basis.phi_0)

        # Can we extract the phase shift faster?
        self.phi_components = np.hstack(( self.basis.phi_0[:, np.newaxis], self.basis.vectors ))
        d1_operator = finite_difference_first_derivative(self.s_mesh)
        self.phi_prime_components = d1_operator @ self.phi_components
    

    def coefficients(self,
        theta: np.array

    ):
        r'''Calculated the coefficients in the $\hat{\phi}$ expansion.

        Parameters:
            theta (ndarray): point in parameters space

        Returns:
            coefficients (ndarray): coefficients in $\hat{\phi}$ expansion

        '''
        invk, beta = self.interaction.coefficients(theta)

        A_utilde = np.einsum('i,ijk', beta, self.A_2)
        A = self.A_13 + A_utilde + invk * self.A_3_coulomb

        b_utilde = beta @ self.b_2
        b = self.b_13 + b_utilde + invk * self.b_3_coulomb

        return np.linalg.solve(A, b)


    def emulate_wave_function(self,
        theta: np.array
    ):
        r'''Calculate the emulated wave function.

        Parameters:
            theta (ndarray): point in parameters space

        Returns:
            wave_function (ndarray): $\hat{\phi}$ on the default $s$ mesh

        '''
        x = self.coefficients(theta)
        return self.basis.phi_hat(x)
    

    def emulate_phase_shift(self,
        theta: np.array
    ):
        r'''Calculate the emulated phase shift.

        Parameters:
            theta (ndarray): point in parameters space
        
        Returns:
            delta (float | complex): emulated phase shift (extracted at $s_0$)

        '''
        x = self.coefficients(theta)
        phi = np.sum(np.hstack((1, x)) * self.phi_components[self.i_0, :])
        phi_prime = np.sum(np.hstack((1, x)) * self.phi_prime_components[self.i_0, :])

        rl = 1/self.s_0 * (phi/phi_prime)
        return np.log(
            (self.Hm - self.s_0*rl*self.Hmp) / 
            (self.Hp - self.s_0*rl*self.Hpp)
        ) / 2j
    
    
    def logarithmic_derivative(self,
        theta: np.array
    ):
        a = self.s_mesh[self.i_0]
        x = self.coefficients(theta)
        phi = np.sum(np.hstack((1, x)) * self.phi_components[self.i_0, :])
        phi_prime = np.sum(np.hstack((1, x)) * self.phi_prime_components[self.i_0, :])
        return 1/a * phi / phi_prime
    
    
    def S_matrix_element(self,
        theta: np.array
    ):
        a = self.s_mesh[self.i_0]
        Rl = self.logarithmic_derivative(theta)
        eta = self.interaction.eta(theta)
        return (self.Hm - a*Rl*self.Hmp) / (self.Hp - a*Rl*self.Hpp)


    def exact_phase_shift(self, theta: np.array):
        r'''Calculate the high-fidelity phase shift.
        
        Parameters:
            theta (ndarray): point in parameters space
        
        Returns:
            delta (float | complex): high-fidelity phase shift (extracted at $s_0$)

        '''
        return self.basis.solver.delta(theta, self.s_mesh[[0, -1]], self.l, self.s_0)
    

    def save(self, filename):
        r'''Write the current emulator to file.
        
        Parameters:
            filename (string): name of the file where the emulator is saved
        
        Returns:
            _ (None): nothing
        '''
        with open(filename, 'wb') as f:
            pickle.dump(self, f)



from .interaction import Interaction, InteractionSpace
from .interaction_eim import InteractionEIM, InteractionEIMSpace
from .energized_interaction_eim import EnergizedInteractionEIM, EnergizedInteractionEIMSpace
from .mn_potential import MN_Potential
from .schroedinger import SchroedingerEquation
from . import constants, metrics
from .reduced_basis_emulator import ReducedBasisEmulator
from .scattering_amplitude_emulator import ScatteringAmplitudeEmulator
from .free_solutions import phase_shift
from .basis import CustomBasis, RelativeBasis
from .koning_delaroche import KoningDelaroche, EnergizedKoningDelaroche
from .spin_orbit import SpinOrbitTerm

from .__version__ import __version__



'''Interactions that leverage the Empirical Interpolation Method (EIM) to allow
the emulation of parameters in which the coordinate-space potential is not
affine.
'''

from typing import Callable
import numpy as np
from scipy.stats import qmc

from .interaction import Interaction, InteractionSpace, couplings
from .constants import HBARC, DEFAULT_RHO_MESH
from .spin_orbit import SpinOrbitTerm

def max_vol(basis, indxGuess):
    r'''basis looks like a long matrix, the columns are the "pillars" V_i(x):
        [   V_1(x)
            V_2(x)
            .
            .
            .
        ]
        indxGuess is a first guess of where we should "measure", or ask the questions

    '''
    nbases = basis.shape[1]
    interpBasis = np.copy(basis)

    for ij in range(len(indxGuess)):
        interpBasis[[ij,indxGuess[ij]],:] = interpBasis[[indxGuess[ij],ij],:]
    indexing = np.array(range(len(interpBasis)))

    for ij in range(len(indxGuess)):
        indexing[[ ij,indxGuess[ij] ]] = indexing[[ indxGuess[ij],ij ]]
    
    for iIn in range(1, 100):
        B = np.dot(interpBasis, np.linalg.inv(interpBasis[:nbases]))
        b = np.max(B)
        if b > 1:
            
            p1, p2 = np.where(B == b)[0][0], np.where(B == b)[1][0]
            interpBasis[[p1,p2],:] = interpBasis[[p2,p1],:]
            indexing[[p1,p2]] = indexing[[p2,p1]]
        else:
            break
        #this thing returns the indexes of where we should measure
    return np.sort(indexing[:nbases])


class InteractionEIM(Interaction):
    def __init__(self,
        coordinate_space_potential: Callable[[float, np.array], float], # V(r, theta)
        n_theta: int, # How many parameters does the interaction have?
        mu: float, # reduced mass (MeV)
        energy: float, # E_{c.m.}
        ell: int,
        training_info: np.array,
        Z_1: int = 0, # atomic number of particle 1
        Z_2: int = 0, # atomic number of particle 2
        R_C: float = 0.0, # Coulomb "cutoff"
        is_complex: bool = False,
        spin_orbit_term: SpinOrbitTerm = None,
        n_basis: int = None,
        explicit_training: bool = False,
        n_train: int = 1000,
        rho_mesh: np.array = DEFAULT_RHO_MESH,
        match_points: np.array = None
    ):
        r'''
        Parameters:
            coordinate_space_potential (Callable[[float,ndarray],float]): V(r,
                theta) where theta are the interaction parameters
            n_theta (int): number of interaction parameters
            mu (float): reduced mass (MeV); converted to 1/fm
            energy (float): center-of-mass scattering energy
            ell (int): angular momentum
            training_info (ndarray): Either (1) parameters bounds or (2)
                explicit training points

                If (1):
                    This is a 2-column matrix. The first column are the lower
                    bounds. The second are the upper bounds. Each row maps to a
                    single parameter.

                If (2):
                    This is an MxN matrix. N is the number of parameters. M is
                    the number of samples.
            Z_1 (int): charge of particle 1
            Z_2 (int): charge of particle 2
            R_C (float): Coulomb "cutoff" radius
            is_complex (bool): Is the interaction complex (e.g. optical
                potentials)?
            spin_orbit_term (SpinOrbitTerm): spin-orbit part of the interaction
            n_basis (int): number of basis states, or "pillars" in $\hat{U}$ approximation
            explicit_training (bool): Is training_info (1) or (2)? (1) is
                default
            n_train (int): How many snapshots to generate? Ignored if
                explicit_training is True.
            rho_mesh (ndarray): coordinate-space points at which the interaction
                is generated (used for training)
            match_points (ndarray): $\rho$ points where agreement with the true
                potential is enforced

        Attributes:
            s_mesh (ndarray): $s$ points
            singular_values (ndarray): `S` in `U, S, Vt = numpy.linalg.svd(...)`
            snapshots (ndarray): pillars, columns of `U`
            match_indices (ndarray): indices of points in $\rho$ mesh that are
                matched to the true potential
            match_points (ndarray): points in $\rho$ mesh that are matched to
                the true potential
            r_i (ndarray): copy of `match_points` (???)
            Ainv (ndarray): inverse of A matrix (Ax = b)
        '''

        super().__init__(coordinate_space_potential, n_theta, mu, energy, ell,
            Z_1=Z_1, Z_2=Z_2, R_C=R_C, is_complex=is_complex, spin_orbit_term=spin_orbit_term)

        self.s_mesh = rho_mesh.copy()
        # Generate a basis used to approximate the potential.
        # Did the user specify the training points?
        if explicit_training:
            snapshots = np.array([self.tilde(rho_mesh, theta) for theta in training_info]).T
        else:
            # Generate training points using the user-provided bounds.
            sampler = qmc.LatinHypercube(d=len(training_info))
            sample = sampler.random(n_train)
            train = qmc.scale(sample, training_info[:, 0], training_info[:, 1])
            snapshots = np.array([self.tilde(rho_mesh, theta) for theta in train]).T
        
        U, S, _ = np.linalg.svd(snapshots, full_matrices=False)
        self.singular_values = np.copy(S)
        
        if match_points is None:
            if n_basis is None:
                n_basis = n_theta
            self.snapshots = np.copy(U[:, :n_basis])
            # random r points between 0 and 2π fm
            i_max = self.snapshots.shape[0] // 2
            di = i_max // (n_basis - 1)
            i_init = np.arange(0, i_max + 1, di)
            self.match_indices = max_vol(self.snapshots, i_init)
                # np.random.randint(0, self.snapshots.shape[0], size=self.snapshots.shape[1]))
            self.match_points = rho_mesh[self.match_indices]
            self.r_i = np.copy(self.match_points)
        else:
            n_basis = match_points.size
            self.snapshots = np.copy(U[:, :n_basis])
            self.match_points = np.copy(match_points)
            self.match_indices = np.array([np.argmin(np.abs(rho_mesh - ri)) for ri in self.match_points])
            self.r_i = rho_mesh[self.match_indices]

        self.Ainv = np.linalg.inv(self.snapshots[self.match_indices])


    def coefficients(self,
        alpha: np.array
    ):
        r'''Computes the EIM expansion coefficients.

        Parameters:
            alpha (ndarray): interaction parameters
        
        Returns:
            coefficients (ndarray): EIM expansion coefficients

        '''
        u_true = self.tilde(self.r_i, alpha)
        return 1/self.k, self.Ainv @ u_true


    def tilde_emu(self,
        alpha: np.array
    ):
        r'''Emulated interaction = $\hat{U}(s, \alpha, E)$

        Parameters:
            alpha (ndarray): interaction parameters
        
        Returns:
            u_hat (ndarray): emulated interaction

        '''
        _, x = self.coefficients(alpha)
        emu = np.sum(x * self.snapshots, axis=1)
        return emu
    

    def basis_functions(self, s_mesh: np.array):
        r'''$u_j$ in $\tilde{U} \approx \hat{U} \equiv \sum_j \beta_j(\alpha) u_j$

        Parameters:
            s_mesh (ndarray): $s$ mesh points
        
        Returns:
            u_j (ndarray): "pillars" (MxN matrix; M = number of mesh points; N = number of pillars)

        '''
        return np.copy(self.snapshots)
    

class InteractionEIMSpace(InteractionSpace):
    def __init__(self,
        coordinate_space_potential: Callable[[float, np.array], float], # V(r, theta)
        n_theta: int, # How many parameters does the interaction have?
        mu: float, # reduced mass (MeV)
        energy: float, # E_{c.m.}
        l_max: int,
        training_info: np.array,
        Z_1: int = 0, # atomic number of particle 1
        Z_2: int = 0, # atomic number of particle 2
        R_C: float = 0.0, # Coulomb "cutoff"
        is_complex: bool = False,
        spin_orbit_potential: Callable[[float, np.array, float], float] = None, #V_{SO}(r, theta, l•s)
        n_basis: int = None,
        explicit_training: bool = False,
        n_train: int = 1000,
        rho_mesh: np.array = DEFAULT_RHO_MESH,
        match_points: np.array = None
    ):
        r'''Generates a list of $\ell$-specific, EIMed interactions.
        
        Parameters:
            coordinate_space_potential (Callable[[float,ndarray],float]): V(r, theta)
            n_theta (int): number of parameters
            mu (float): reduced mass
            energy (float): center-of-mass, scattering energy
            l_max (int): maximum angular momentum
            training_info (ndarray): See `InteractionEIM` documentation.
            Z_1 (int): charge of particle 1
            Z_2 (int): charge of particle 2
            R_C (float): Coulomb "cutoff" radius
            is_complex (bool): Is the interaction complex?
            spin_orbit_potential (Callable[[float, np.array, float], float]):
                used to create a `SpinOrbitTerm`
            n_basis (int): number of pillars --- basis states in $\hat{U}$ expansion
            explicit_training (bool): See `InteractionEIM` documentation.
            n_train (int): number of training samples
            rho_mesh (ndarray): discrete $\rho$ points
            match_points (ndarray): $\rho$ points where agreement with the true
                potential is enforced
        
        Returns:
            instance (InteractionEIMSpace): instance of InteractionEIMSpace
        
        Attributes:
            interactions (list): list of `InteractionEIM`s
        '''
        self.interactions = []
        if spin_orbit_potential is None:
            for l in range(l_max+1):
                self.interactions.append(
                    [InteractionEIM(coordinate_space_potential, n_theta, mu,
                        energy, l, training_info, Z_1=Z_1, Z_2=Z_2, R_C=R_C,
                        is_complex=is_complex, n_basis=n_basis, explicit_training=explicit_training,
                        n_train=n_train, rho_mesh=rho_mesh, match_points=match_points)]
                )
        else:
            for l in range(l_max+1):
                self.interactions.append(
                    [InteractionEIM(coordinate_space_potential, n_theta, mu,
                        energy, l, training_info, Z_1=Z_1, Z_2=Z_2, R_C=R_C,
                        is_complex=is_complex,
                        spin_orbit_term=SpinOrbitTerm(spin_orbit_potential,
                        lds), n_basis=n_basis,
                        explicit_training=explicit_training, n_train=n_train,
                        rho_mesh=rho_mesh, match_points=match_points) for lds in
                        couplings(l)]
                )


import pickle
import numpy as np
from scipy.special import eval_legendre, gamma
from tqdm import tqdm 

from .interaction import InteractionSpace
from .reduced_basis_emulator import ReducedBasisEmulator
from .constants import DEFAULT_RHO_MESH, DEFAULT_ANGLE_MESH
from .schroedinger import SchroedingerEquation
from .basis import RelativeBasis, CustomBasis
from .utility import eval_assoc_legendre

class ScatteringAmplitudeEmulator:

    @classmethod
    def load(obj, filename):
        r'''Loads a previously trained emulator.

        Parameters:
            filename (string): name of file
        
        Returns:
            emulator (ScatteringAmplitudeEmulator): previously trainined `ScatteringAmplitudeEmulator`
        
        '''
        with open(filename, 'rb') as f:
            sae = pickle.load(f)
        return sae


    @classmethod
    def from_train(cls,
        interaction_space: InteractionSpace,
        theta_train: np.array,
        l_max: int,
        angles: np.array = DEFAULT_ANGLE_MESH,
        n_basis: int = 4,
        use_svd: bool = True,
        s_mesh: np.array = DEFAULT_RHO_MESH,
        s_0: float = 6*np.pi,
        hf_tols: list = None
    ):
        r'''Trains a reduced-basis emulator based on the provided interaction and training space.
        
        Parameters:
            interaction_space (InteractionSpace): local interaction up to (and including $\ell_\max$)
            theta_train (ndarray): training points in parameter space; shape = (n_points, n_parameters)
            l_max (int): maximum angular momentum to include in the sum approximating the cross section
            angles (ndarray): Differential cross sections are functions of the
                angles. These are the specific values at which the user wants to
                emulate the cross section.
            n_basis (int): number of basis vectors for $\hat{\phi}$ expansion
            use_svd (bool): Use principal components of training wave functions?
            s_mesh (ndarray): $s$ (or $\rho$) grid on which wave functions are evaluated
            s_0 (float): $s$ point where the phase shift is extracted
            hf_tols (list): 2-element list passed to `SchroedingerEquation`;
                [relative tolerance, absolute tolerance]
        
        Returns:
            sae (ScatteringAmplitudeEmulator): scattering amplitude emulator
        
        '''
        bases = []
        for interaction_list in tqdm(interaction_space.interactions):
            basis_list = [RelativeBasis(
                SchroedingerEquation(interaction, hifi_tolerances=hf_tols),
                theta_train, s_mesh, n_basis, interaction.ell, use_svd
            ) for interaction in interaction_list]
            bases.append(basis_list)

        return cls(interaction_space, bases, l_max, angles=angles, s_0=s_0)


    def __init__(self,
        interaction_space: InteractionSpace,
        bases: list,
        l_max: int,
        angles: np.array = DEFAULT_ANGLE_MESH,
        s_0: float = 6*np.pi,
        verbose: bool = True
    ):
        r'''Trains a reduced-basis emulator that computes differential and total cross sections (from emulated phase shifts).

        Parameters:
            interaction_space (InteractionSpace): local interaction up to (and including $\ell_\max$)
            bases (list[Basis]): list of `Basis` objects
            l_max (int): maximum angular momentum to include in the sum approximating the cross section
            angles (ndarray): Differential cross sections are functions of the
                angles. These are the specific values at which the user wants to
                emulate the cross section.
            s_0 (float): $s$ point where the phase shift is extracted
            verbose (bool): Do you want the class to print out warnings?

        Attributes:
            l_max (int): maximum angular momentum
            angles (ndarray): angle values at which the differential cross section is desired
            rbes (list): list of `ReducedBasisEmulators`; one for each partial wave (and total $j$ with spin-orbit)
            ls (ndarray): angular momenta; shape = (`l_max`+1, 1)
            P_l_costheta (ndarray): Legendre polynomials evaluated at `angles`
            P_1_l_costheta (ndarray): **associated** Legendre polynomials evalated at `angles`
            k_c (float): Coulomb momentum, $k\eta$
            eta (float): Sommerfeld parameter
            sigma_l (float): Coulomb phase shift
            f_c (ndarray): scattering amplitude
            rutherford (ndarray): Rutherford scattering

        '''
        self.l_max = l_max
        self.angles = angles.copy()
        bases_types = [isinstance(bases[l], CustomBasis) for l in range(self.l_max+1)]
        if np.any(bases_types) and verbose:
            print('''
NOTE: When supplying a CustomBasis, the ROSE high-fidelity solver is \n
instantiated for the sake of future evaluations.  Any requests to the solver \n
will NOT be communicated to the user's own high-fidelity solver.
''')
        self.rbes = []
        for (interaction_list, basis_list) in zip(interaction_space.interactions, bases):
            self.rbes.append([ReducedBasisEmulator(interaction, basis, s_0=s_0) for
                (interaction, basis) in zip(interaction_list, basis_list)])


        # Let's precompute the things we can.
        self.ls = np.arange(self.l_max+1)[:, np.newaxis]
        self.P_l_costheta = eval_legendre(self.ls, np.cos(self.angles))
        self.P_1_l_costheta = np.array([eval_assoc_legendre(l, np.cos(self.angles)) for l in self.ls])
        # Coulomb scattering amplitude
        # (This is dangerous because it's not fixed when we emulate across
        # energies, BUT we don't do that with Coulomb (yet). When we do emulate
        # across energies, f_c is zero anyway.)
        k = self.rbes[0][0].interaction.momentum(None)
        self.k_c = self.rbes[0][0].interaction.k_c
        self.eta = self.k_c / k
        self.sigma_l = np.angle(gamma(1 + self.ls + 1j*self.eta))
        sin2 = np.sin(self.angles/2)**2
        self.f_c = -self.eta / (2*k*sin2) * np.exp(-1j*self.eta*np.log(sin2) + 2j*self.sigma_l[0])
        self.rutherford = self.eta**2 / (4*k**2*np.sin(self.angles/2)**4)


    def emulate_phase_shifts(self,
        theta: np.array
    ):
        r'''Gives the phase shifts for each partial wave.  Order is [l=0, l=1,
            ..., l=l_max-1].
        
        Parameters:
            theta (ndarray): parameter-space vector
        
        Returns:
            phase_shift (list): emulated phase shifts

        '''
        return [[rbe.emulate_phase_shift(theta) for rbe in rbe_list] for rbe_list in self.rbes]

    def exact_phase_shifts(self,
        theta: np.array
    ):
        r'''Gives the phase shifts for each partial wave. Order is [l=0, l=1,
            ..., l=l_max-1].
        
        Parameters:
            theta (ndarray): parameter-space vector
        
        Returns:
            phase_shift (list): high-fidelity phase shifts

        '''
        return [[rbe.exact_phase_shift(theta) for rbe in rbe_list] for rbe_list in self.rbes]


    def dsdo(self, theta : np.array, deltas : np.array):
        r'''Gives the differential cross section (dsigma/dOmega = dsdo).

        Parameters:
            theta (ndarray): parameter-space vector
            deltas (ndarray): phase shifts
        
        Returns:
            dsdo (ndarray): differential cross section (fm^2)

        '''
        k = self.rbes[0][0].interaction.momentum(theta)

        S_l_plus, S_l_minus = self.S_matrix_elements(deltas)

        A = self.f_c + 1/(2j*k) * np.sum(
            np.exp(2j*self.sigma_l) * ((self.ls+1)*(S_l_plus - 1) + \
                self.ls*(S_l_minus - 1)) * self.P_l_costheta,
            axis=0
        )
        B = 1/(2j*k) * np.sum(
            np.exp(2j*self.sigma_l) * (S_l_plus - S_l_minus) * self.P_1_l_costheta,
            axis=0
        )

        dsdo = (np.conj(A)*A + np.conj(B)*B).real
        if self.k_c > 0:
            return dsdo / self.rutherford
        else:
            return dsdo


    def emulate_dsdo(self,
        theta: np.array
    ):
        r'''Emulates the differential cross section (dsigma/dOmega = dsdo).

        Parameters:
            theta (ndarray): parameter-space vector

        Returns:
            dsdo (ndarray): emulated differential cross section

        '''
        deltas = self.emulate_phase_shifts(theta)
        return self.dsdo(theta, deltas)


    def exact_dsdo(self,
        theta: np.array
    ):
        r'''Calculates the high-fidelity differential cross section (dsigma/dOmega = dsdo).

        Parameters:
            theta (ndarray): parameter-space vector

        Returns:
            dsdo (ndarray): high-fidelity differential cross section

        '''
        deltas = self.exact_phase_shifts(theta)
        return self.dsdo(theta, deltas)


    def emulate_wave_functions(self,
        theta: np.array
    ):
        r'''Gives the wave functions for each partial wave.  Returns a list of
            arrays.  Order is [l=0, l=1, ..., l=l_max-1].
        
        Parameters:
            theta (ndarray): parameter-space vector

        Returns:
            wave_functions (list): emulated wave functions

        
        '''
        return [[x.emulate_wave_function(theta) for x in rbe] for rbe in self.rbes]


    def S_matrix_elements(self,
        deltas: list
    ):
        deltas_plus = np.array([d[0] for d in deltas])
        deltas_minus = np.array([d[1] for d in deltas[1:]])

        S_l_plus = np.exp(2j*deltas_plus)[:, np.newaxis]
        if self.rbes[0][0].interaction.include_spin_orbit:
            # If there is spin-orbit, the l=0 term for B has to be zero.
            S_l_minus = np.hstack((S_l_plus[0], np.exp(2j*deltas_minus)))[:, np.newaxis]
        else:
            # This ensures that A reduces to the non-spin-orbit formula, and B = 0.
            S_l_minus = S_l_plus.copy()
        
        return S_l_plus, S_l_minus


    def emulate_total_cross_section(self,
        theta: np.array
    ):
        r'''Gives the "total" (angle-integrated) cross section.  See Eq. (63)
            in Carlson's notes.
        
        Parameters:
            theta (ndarray): parameter-space vector
        
        Returns:
            cross_section (ndarray): emulated total cross section

        '''
        # What do we do here when Coulomb and/or spin-orbit is present?
        if self.k_c > 0:
            raise Exception('The total cross section is infinite in the presence of Coulomb.')
        
        k = self.rbes[0][0].interaction.momentum(theta)
        S_l_plus, S_l_minus = self.S_matrix_elements(self.emulate_phase_shifts(theta))

        sum = np.sum(np.pi/k**2 * (2*self.ls + 2) * (1 - S_l_plus.real))
        sum += np.sum(np.pi/k**2 * (2*self.ls - 2) * (1 - S_l_minus.real))

        return sum


    def exact_total_cross_section(self,
        theta: np.array
    ):
        r'''Gives the "total" (angle-integrated) cross section.  See Eq. (63)
            in Carlson's notes.
        
        Parameters:
            theta (ndarray): parameter-space vector
        
        Returns:
            cross_section (ndarray): emulated total cross section

        '''
        # What do we do here when Coulomb and/or spin-orbit is present?
        if self.k_c > 0:
            raise Exception('The total cross section is infinite in the presence of Coulomb.')
        
        k = self.rbes[0][0].interaction.momentum(theta)
        S_l_plus, S_l_minus = self.S_matrix_elements(self.exact_phase_shifts(theta))

        sum = np.sum(np.pi/k**2 * (2*self.ls + 2) * (1 - S_l_plus.real))
        sum += np.sum(np.pi/k**2 * (2*self.ls - 2) * (1 - S_l_minus.real))

        return sum


    def save(self, filename):
        r'''Saves the emulator to the desired file.
        
        Parameters:
            filename (string): name of file
        
        '''
        with open(filename, 'wb') as f:
            pickle.dump(self, f)


r'''`SchroedingerEquation` is a high-fidelity (HF), Schrödinger-equation solver for
local, complex interactions.

By default, `rose` will provide HF solution using `scipy.integrate.solve_ivp`.
For details about providing your own solutions, see [Basis
documentation](basis.md).

'''
import numpy as np
from scipy.integrate import solve_ivp

from .interaction import Interaction
from .free_solutions import H_minus, H_plus, H_minus_prime, H_plus_prime
from .utility import regular_inverse_s

# Default values for solving the SE.
DEFAULT_R_MIN = 1e-12 # fm
DEFAULT_R_MAX = 30.0 # fm
DEFAULT_R_0 = 20.0 # fm
DEFAULT_NUM_PTS = 2000
MAX_STEPS = 20000
PHI_THRESHOLD = 1e-10

class SchroedingerEquation:
    '''
    High-fidelity (HF) solver for optical potentials. How high-fidelity? You decide!
    '''
    def __init__(self,
        interaction: Interaction,
        hifi_tolerances: list = [1e-12, 1e-12]
    ):
        r'''Solves the Shrödinger equation for local, complex potentials.
        
        Parameters:
            interaction (Interaction): See [Interaction documentation](interaction.md).
            hifi_tolerances (list): 2-element list of numbers: the relative
                tolerance, `rel_tol`, and the absolute tolerance `abs_tol`

        Returns:
            solver (SchroedingerEquation): instance of `SchroedingerEquation`

        '''
        self.interaction = interaction
        self.rel_tol = hifi_tolerances[0]
        self.abs_tol = hifi_tolerances[1]


    def solve_se(self,
        alpha: np.array, # interaction parameters
        s_endpts: np.array, # s where phi(s) is calculated
        l: int = 0, # angular momentum
        rho_0 = None, # initial rho value ("effective zero")
        phi_threshold = PHI_THRESHOLD, # minimum phi value (zero below this value)
        **solve_ivp_kwargs
    ):
        r'''Solves the reduced, radial Schrödinger equation.
        
        Parameters:
            alpha (ndarray): parameter vector
            s_endpts (ndarray): lower and upper bounds of the $s$ mesh.
            l (int): angular momentum
            rho_0 (float): initial $\rho$ (or $s$) value; starting point for the
                solver
            phi_threshold (float): minimum $\phi$ value; The wave function is
                considered zero below this value.
        
        Returns:
            array (ndarray): 2-column matrix; The first column is the $r$ values.
                The second is the reduced radial wavefunction, $u(r)$. (The optional
                third - based on return_uprime - is $u^\prime(r)$.)
        '''

        C_l = Gamow_factor(l, self.interaction.eta(alpha))
        S_C = self.interaction.momentum(alpha) * self.interaction.R_C

        if rho_0 is None:
            rho_0 = (phi_threshold / C_l) ** (1/(l+1))
        phi_0 = C_l * rho_0**(l+1)
        phi_prime_0 = C_l * (l+1) * rho_0**l
        
        if self.interaction.is_complex:
            initial_conditions = np.array([phi_0+0j, phi_prime_0+0j])
        else:
            initial_conditions = np.array([phi_0, phi_prime_0])
        
        sol = solve_ivp(
            lambda s, phi: np.array([phi[1],
                (self.interaction.tilde(s, alpha) + \
                 2*self.interaction.eta(alpha) * regular_inverse_s(s, S_C) + l*(l+1)/s**2 - 1.0) * phi[0]]),
            s_endpts, initial_conditions, rtol=self.rel_tol, atol=self.abs_tol,
            dense_output=True, **solve_ivp_kwargs
        )

        return sol.sol


    def delta(self,
        alpha: np.array, # interaction parameters
        s_endpts: np.array, # [s_min, s_max]; phi(s) is calculated on this interval 
        l: int, # angular momentum
        s_0: float, # phaseshift is extracted at phi(s_0)
        **solve_ivp_kwargs # passed to solve_se
    ):
        r'''Calculates the $\ell$-th partial wave phase shift at the specified energy.
        solve_ivp_kwargs are passed to solve_se

        Parameters:
            alpha (ndarray): parameter vector
            s_endpts (ndarray): lower and upper bounds of the $s$ mesh.
            l (int): angular momentum
            s_0 (float): $s$ value where the phase shift is calculated (must be
                less than the second element in `s_endpts`)
        
        Returns:
            delta (float): phase shift extracted from the reduced, radial
                wave function

        '''
        # Should s_endpts be [s_min, s_endpts[1]]?
        solution = self.solve_se(alpha, s_endpts, l=l, **solve_ivp_kwargs)
        u = solution(s_0)
        rl = 1/s_0 * (u[0]/u[1])
        return np.log(
            (H_minus(s_0, l, self.interaction.eta(alpha)) - s_0*rl*H_minus_prime(s_0, l, self.interaction.eta(alpha))) / 
            (H_plus(s_0, l, self.interaction.eta(alpha)) - s_0*rl*H_plus_prime(s_0, l, self.interaction.eta(alpha)))
        ) / 2j


    def phi(self,
        alpha: np.array, # interaction parameters
        s_mesh: np.array, # s where phi(s) in calculated
        l: int, # angular momentum
        rho_0: float = None, # What do we call "zero"?
        phi_threshold: float = PHI_THRESHOLD,
        **solve_ivp_kwargs # passed to solve_se
    ):
        r'''Computes the reduced, radial wave function $\phi$ (or $u$) on `s_mesh`.

        Parameters:
            alpha (ndarray): parameter vector
            s_mesh (ndarray): values of $s$ at which $\phi$ is calculated
            l (int): angular momentum
            rho_0 (float): starting point for the solver
            phi_threshold (float): minimum $\phi$ value; The wave function is
                considered zero below this value.
        
        Returns:
            phi (ndarray): reduced, radial wave function

        '''
        if rho_0 is None:
            rho_0 = (phi_threshold / Gamow_factor(l, self.interaction.eta(alpha))) ** (1/(l+1))

        solution = self.solve_se(alpha, [rho_0, s_mesh[-1]], l, rho_0=rho_0,
                                 phi_threshold=phi_threshold, **solve_ivp_kwargs)

        ii_0 = np.where(s_mesh < rho_0)[0]
        y = solution(s_mesh)[0]
        y[ii_0] = 0
        return y
    

def Gamow_factor(l, eta):
    r'''This returns the... Gamow factor.
    See [Wikipedia](https://en.wikipedia.org/wiki/Gamow_factor).

    Parameters:
        l (int): angular momentum
        eta (float): Sommerfeld parameter (see
            [Wikipedia](https://en.wikipedia.org/wiki/Sommerfeld_parameter))
    
    Returns:
        C_l (float): Gamow factor

    '''
    if eta == 0.0:
        if l == 0:
            return 1
        else:
            return 1 / (2*l + 1) * Gamow_factor(l-1, 0)
    elif l == 0:
        return np.sqrt(2*np.pi*eta / (np.exp(2*np.pi*eta)-1))
    else:
        return np.sqrt(l**2 + eta**2) / (l*(2*l+1)) * Gamow_factor(l-1, eta)



'''Minnesota potentials for testing.
'''

import numpy as np
from .interaction import Interaction

NUCLEON_MASS = 939.565 # neutron mass (MeV)
MU_NN = NUCLEON_MASS / 2 # reduced mass of the NN system (MeV)

def mn_potential(r, args):
    '''
    Minnesota potential
    '''
    v_0r, v_0s = args
    return v_0r * np.exp(-1.487*r**2) + v_0s*np.exp(-0.465*r**2)

# Stored instances of the Minnesota interaction for testing.
# Fixed at E_{c.m.} = 50 MeV.
MN_Potential = Interaction(
    mn_potential,
    2,
    MU_NN,
    50,
    0
)

def complex_mn_potential(r, args):
    vr, vi = args
    return mn_potential(r, [vr, 1j*vi])


Complex_MN_Potential = Interaction(
    complex_mn_potential,
    2,
    MU_NN,
    50,
    0,
    is_complex = True
)



'''
Functions for computing metrics.
'''

import pickle

import numpy as np

from .reduced_basis_emulator import ReducedBasisEmulator
from .free_solutions import phase_shift_interp

def wave_function_metric(
    rbe: ReducedBasisEmulator,
    filename: str
):
    with open(filename, 'rb') as f:
        benchmark_data = pickle.load(f)
    
    thetas = np.array([bd.theta for bd in benchmark_data])
    wave_functions = np.array([bd.phi for bd in benchmark_data])
    emulated_wave_functions = np.array([rbe.emulate_wave_function(theta) for theta in thetas])
    abs_residuals = np.abs(emulated_wave_functions - wave_functions)
    norm = np.sqrt(np.sum(abs_residuals**2, axis=1))
    return np.quantile(norm, [0.5, 0.95])


def phase_shift_metric(
    rbe: ReducedBasisEmulator,
    filename: str
):
    with open(filename, 'rb') as f:
        benchmark_data = pickle.load(f)
    
    thetas = np.array([bd.theta for bd in benchmark_data])

    wave_functions = np.array([bd.phi for bd in benchmark_data])
    phase_shifts = np.array([phase_shift_interp(u, rbe.s_mesh, rbe.l, rbe.interaction.eta, rbe.s_0) for u in wave_functions])

    emulated_phase_shifts = np.array([rbe.emulate_phase_shift(theta) for theta in thetas])

    rel_diff = np.abs((emulated_phase_shifts - phase_shifts) / emulated_phase_shifts)
    med, upper = np.quantile(rel_diff, [0.5, 0.95])
    return [med, upper, np.max(np.abs(rel_diff))]


def run_metrics(
    rbe: ReducedBasisEmulator,
    filename: str,
    verbose: bool = False
):
    wave_function_results = wave_function_metric(rbe, filename)
    phase_shift_results = phase_shift_metric(rbe, filename)

    if verbose:
        print('Wave function residuals (root of sum of squares):\n50% and 95% quantiles')
        print(f'{wave_function_results[0]:.4e}  {wave_function_results[1]:.4e}')
        print('Phase shift residuals (relative difference):\n50% and 95% quantiles')
        print(f'{phase_shift_results[0]:.4e}  {phase_shift_results[1]:.4e}')
        print('Maximum phase shift residuals')
        print(f'{phase_shift_results[2]:.4e}')

    return wave_function_results, phase_shift_results


'''
Defines a class for "affinizing" Interactions using the Empirical Interpolation
Method (EIM).
'''

from typing import Callable
import numpy as np
from scipy.stats import qmc

from .interaction import Interaction, InteractionSpace, couplings
from .interaction_eim import InteractionEIM, max_vol
from .constants import HBARC, DEFAULT_RHO_MESH
from .spin_orbit import SpinOrbitTerm

class EnergizedInteractionEIM(Interaction):
    def __init__(self,
        coordinate_space_potential: Callable[[float, np.array], float], # V(r, theta)
        n_theta: int, # How many parameters does the interaction have?
        mu: float, # reduced mass (MeV)
        ell: int,
        training_info: np.array,
        Z_1: int = 0, # atomic number of particle 1
        Z_2: int = 0, # atomic number of particle 2
        R_C: float = 0.0, # Coulomb "cutoff"
        is_complex: bool = False,
        spin_orbit_term: SpinOrbitTerm = None,
        n_basis: int = None,
        explicit_training: bool = False,
        n_train: int = 1000,
        rho_mesh: np.array = DEFAULT_RHO_MESH,
        match_points: np.array = None
    ):
        r'''
        Parameters:
            coordinate_space_potential (Callable[[float,ndarray],float]): V(r,
                theta) where theta are the interaction parameters
            n_theta (int): number of interaction parameters
            mu (float): reduced mass (MeV); converted to 1/fm
            ell (int): angular momentum
            training_info (ndarray): Either (1) parameters bounds or (2)
                explicit training points

                If (1):
                    This is a 2-column matrix. The first column are the lower
                    bounds. The second are the upper bounds. Each row maps to a
                    single parameter.

                If (2):
                    This is an MxN matrix. N is the number of parameters. M is
                    the number of samples.
            Z_1 (int): charge of particle 1
            Z_2 (int): charge of particle 2
            R_C (float): Coulomb "cutoff" radius
            is_complex (bool): Is the interaction complex (e.g. optical
                potentials)?
            spin_orbit_term (SpinOrbitTerm): spin-orbit part of the interaction
            n_basis (int): number of basis states, or "pillars" in $\hat{U}$ approximation
            explicit_training (bool): Is training_info (1) or (2)? (1) is
                default
            n_train (int): How many snapshots to generate? Ignored if
                explicit_training is True.
            rho_mesh (ndarray): coordinate-space points at which the interaction
                is generated (used for training)
            match_points (ndarray): $\rho$ points where agreement with the true
                potential is enforced

        Attributes:
            singular_values (ndarray): `S` in `U, S, Vt = numpy.linalg.svd(...)`
            snapshots (ndarray): pillars, columns of `U`
            match_indices (ndarray): indices of points in $\rho$ mesh that are
                matched to the true potential
            match_points (ndarray): points in $\rho$ mesh that are matched to
                the true potential
            r_i (ndarray): copy of `match_points` (???)
            Ainv (ndarray): inverse of A matrix (Ax = b)
        '''
        super().__init__(coordinate_space_potential, n_theta, mu, None, ell,
            Z_1=Z_1, Z_2=Z_2, R_C=R_C, is_complex=is_complex, spin_orbit_term=spin_orbit_term)

        # Generate a basis used to approximate the potential.
        # Did the user specify the training points?
        if explicit_training:
            snapshots = np.array([self.tilde(rho_mesh, theta) for theta in training_info]).T
        else:
            # Generate training points using the user-provided bounds.
            sampler = qmc.LatinHypercube(d=len(training_info))
            sample = sampler.random(n_train)
            train = qmc.scale(sample, training_info[:, 0], training_info[:, 1])
            snapshots = np.array([self.tilde(rho_mesh, theta) for theta in train]).T
        
        U, S, _ = np.linalg.svd(snapshots, full_matrices=False)
        self.singular_values = np.copy(S)
        
        if match_points is None:
            if n_basis is None:
                n_basis = n_theta
            self.snapshots = np.copy(U[:, :n_basis])
            # random r points between 0 and 2π fm
            i_max = self.snapshots.shape[0] // 4
            di = i_max // (n_basis - 1)
            i_init = np.arange(0, i_max + 1, di)
            self.match_indices = max_vol(self.snapshots, i_init)
                # np.random.randint(0, self.snapshots.shape[0], size=self.snapshots.shape[1]))
            self.match_points = rho_mesh[self.match_indices]
            self.r_i = np.copy(self.match_points)
        else:
            n_basis = match_points.size
            self.snapshots = np.copy(U[:, :n_basis])
            self.match_points = np.copy(match_points)
            self.match_indices = np.array([np.argmin(np.abs(rho_mesh - ri)) for ri in self.match_points])
            self.r_i = rho_mesh[self.match_indices]

        self.Ainv = np.linalg.inv(self.snapshots[self.match_indices])


    def tilde(self,
        s: float,
        alpha: np.array
    ):
        r'''Computes the energy-scaled interaction.
        
        Parameters:
            s (float): mesh point
            alpha (ndarray): interaction parameters
        
        Returns:
            u_tilde (float | complex): energy-scaled interaction

        '''
        energy = alpha[0]
        k = np.sqrt(2*self.mu*energy/HBARC)
        return  1.0/energy * self.v_r(s/k, alpha[1:])


    def coefficients(self,
        alpha: np.array # interaction parameters
    ):
        r'''Computes the EIM expansion coefficients.

        Parameters:
            alpha (ndarray): interaction parameters
        
        Returns:
            coefficients (ndarray): EIM expansion coefficients

        '''
        k = np.sqrt(2*self.mu*alpha[0]/HBARC)
        u_true = self.tilde(self.r_i, alpha)
        return 1/k, self.Ainv @ u_true


    def eta(self,
        alpha: np.array
    ):
        r'''Returns the Sommerfeld parameter.
        
        Parameters:
            alpha (ndarray): interaction parameters
        
        Returns:
            eta (float): Sommerfeld parameter
        
        '''
        return self.k_c / np.sqrt(2*self.mu*alpha[0]/HBARC)
    

    def tilde_emu(self,
        s: float,
        alpha: np.array
    ):
        r'''Emulated interaction = $\hat{U}(s, \alpha, E)$

        Parameters:
            alpha (ndarray): interaction parameters
        
        Returns:
            u_hat (ndarray): emulated interaction

        '''
        _, x = self.coefficients(alpha)
        emu = np.sum(x * self.snapshots, axis=1)
        return emu
    

    def basis_functions(self, s_mesh: np.array):
        r'''$u_j$ in $\tilde{U} \approx \hat{U} \equiv \sum_j \beta_j(\alpha) u_j$

        Parameters:
            s_mesh (ndarray): $s$ mesh points
        
        Returns:
            u_j (ndarray): "pillars" (MxN matrix; M = number of mesh points; N = number of pillars)

        '''
        return np.copy(self.snapshots)
    

    def momentum(self, alpha: np.array):
        r'''Center-of-mass, scattering momentum
        
        Parameters:
            alpha (ndarray): interaction parameters
        
        Returns:
            k (float): momentum
        '''
        return np.sqrt(2*self.mu*alpha[0]/HBARC)


class EnergizedInteractionEIMSpace(InteractionSpace):
    def __init__(self,
        coordinate_space_potential: Callable[[float, np.array], float], # V(r, theta)
        n_theta: int, # How many parameters does the interaction have?
        mu: float, # reduced mass (MeV)
        l_max: int,
        training_info: np.array,
        Z_1: int = 0, # atomic number of particle 1
        Z_2: int = 0, # atomic number of particle 2
        R_C: float = 0.0, # Coulomb "cutoff"
        is_complex: bool = False,
        spin_orbit_potential: Callable[[float, np.array, float], float] = None, #V_{SO}(r, theta, l•s)
        n_basis: int = None,
        explicit_training: bool = False,
        n_train: int = 1000,
        rho_mesh: np.array = DEFAULT_RHO_MESH,
        match_points: np.array = None
    ):
        r'''Generates a list of $\ell$-specific, energy-emulated, EIMed interactions.
        
        Parameters:
            coordinate_space_potential (Callable[[float,ndarray],float]): V(r, theta)
            n_theta (int): number of parameters
            mu (float): reduced mass
            l_max (int): maximum angular momentum
            training_info (ndarray): See `InteractionEIM` documentation.
            Z_1 (int): charge of particle 1
            Z_2 (int): charge of particle 2
            R_C (float): Coulomb "cutoff" radius
            is_complex (bool): Is the interaction complex?
            spin_orbit_potential (Callable[[float, np.array, float], float]):
                used to create a `SpinOrbitTerm`
            n_basis (int): number of pillars --- basis states in $\hat{U}$ expansion
            explicit_training (bool): See `InteractionEIM` documentation.
            n_train (int): number of training samples
            rho_mesh (ndarray): discrete $\rho$ points
            match_points (ndarray): $\rho$ points where agreement with the true
                potential is enforced
        
        Returns:
            instance (InteractionEIMSpace): instance of InteractionEIMSpace
        
        Attributes:
            interaction (list): list of `InteractionEIM`s
        '''
        self.interactions = []
        if spin_orbit_potential is None:
            for l in range(l_max+1):
                self.interactions.append(
                    [EnergizedInteractionEIM(coordinate_space_potential, n_theta, mu,
                        l, training_info, Z_1=Z_1, Z_2=Z_2, R_C=R_C,
                        is_complex=is_complex, n_basis=n_basis, explicit_training=explicit_training,
                        n_train=n_train, rho_mesh=rho_mesh, match_points=match_points)]
                )
        else:
            for l in range(l_max+1):
                self.interactions.append(
                    [EnergizedInteractionEIM(coordinate_space_potential, n_theta, mu,
                        l, training_info, Z_1=Z_1, Z_2=Z_2, R_C=R_C,
                        is_complex=is_complex,
                        spin_orbit_term=SpinOrbitTerm(spin_orbit_potential,
                        lds), n_basis=n_basis,
                        explicit_training=explicit_training, n_train=n_train,
                        rho_mesh=rho_mesh, match_points=match_points) for lds in
                        couplings(l)]
                )


'''
Useful functions related to the solution to the free, radial Schrödinger equation.
'''

import numpy as np
from scipy.special import spherical_jn, spherical_yn
from mpmath import coulombf, coulombg
from scipy.interpolate import interp1d
from scipy.misc import derivative

def F(rho, ell, eta):
    '''
    Bessel function of the first kind.
    '''
    # return rho*spherical_jn(ell, rho)
    return np.complex128(coulombf(ell, eta, rho))


def G(rho, ell, eta):
    '''
    Bessel function of the second kind.
    '''
    # return -rho*spherical_yn(ell, rho)
    return np.complex128(coulombg(ell, eta, rho))


def H_plus(rho, ell, eta):
    '''
    Hankel function of the first kind.
    '''
    return G(rho, ell, eta) + 1j*F(rho, ell, eta)


def H_minus(rho, ell, eta):
    '''
    Hankel function of the second kind.
    '''
    return G(rho, ell, eta) - 1j*F(rho, ell, eta)


def H_plus_prime(rho, ell, eta, dx=1e-6):
    '''
    Derivative of the Hankel function (first kind) with respect to rho.
    '''
    return derivative(lambda z: H_plus(z, ell, eta), rho, dx=dx)


def H_minus_prime(rho, ell, eta, dx=1e-6):
    '''
    Derivative of the Hankel function (second kind) with respect to rho.
    '''
    return derivative(lambda z: H_minus(z, ell, eta), rho, dx=dx)


def phi_free(rho, ell, eta):
    '''
    Solution to the "free" (V = 0) radial Schrödinger equation.
    '''
    return -0.5j * (H_plus(rho, ell, eta) - H_minus(rho, ell, eta))


# def phase_shift(u, up, ell, x0):
#     rl = 1/x0 * (u/up)
#     return np.log(
#         (H_minus(x0, ell) - x0*rl*H_minus_prime(x0, ell)) / 
#         (H_plus(x0, ell) - x0*rl*H_plus_prime(x0, ell))
#     ) / 2j


def phase_shift(phi, phi_prime, ell, eta, x0):
    rl = 1/x0 * (phi/phi_prime)
    return np.log(
        (H_minus(x0, ell, eta) - x0*rl*H_minus_prime(x0, ell, eta)) / 
        (H_plus(x0, ell, eta) - x0*rl*H_plus_prime(x0, ell, eta))
    ) / 2j


def phase_shift_interp(u, s, ell, eta, x0, dx=1e-6):
    '''
    Given the solution, u, on the s grid, return the phase shift (with respect to the free solution).
    '''
    u_func = interp1d(s, u, kind='cubic')
    rl = 1/x0 * (u_func(x0)/derivative(u_func, x0, dx=dx))
    return np.log(
        (H_minus(x0, ell, eta) - x0*rl*H_minus_prime(x0, ell, eta)) / 
        (H_plus(x0, ell, eta) - x0*rl*H_plus_prime(x0, ell, eta))
    ) / 2j


#!/usr/bin/env python
# coding: utf-8

# In[1]:


import pickle

import numpy as np
from scipy.stats import qmc
from scipy.special import eval_legendre

import rose

from benchmark import BenchmarkData


# In[2]:


sampler = qmc.LatinHypercube(2)
n_samples = 50
sample = sampler.random(n_samples)


# In[3]:


lower_bounds = [100, -200]
upper_bounds = [300, 0]

parameter_samples = qmc.scale(sample, lower_bounds, upper_bounds)


# In[4]:


se = rose.SchroedingerEquation(rose.MN_Potential)


# In[5]:


energy = 50 # MeV
k = np.sqrt(2*rose.MN_Potential.mu*energy/rose.constants.HBARC)
r0s = np.linspace(10.0, 20.0, 11)

phase_shifts = np.array(
    [[se.delta(energy, theta, np.array([k*1e-6, k*30]), 0, s_0=k*r0) for r0 in r0s] for theta in parameter_samples]
)

sigma = np.std(phase_shifts, axis=1)
mu = np.median(phase_shifts, axis=1)
spread = np.max(np.abs(sigma/mu))
print(spread)


# In[6]:


benchmark_data = [BenchmarkData(se, energy, theta) for theta in parameter_samples]


# In[7]:


filename = f'benchmark_data_MN_Potential_E_{energy:.2f}_MeV.pkl'
with open(filename, 'wb') as f:
    pickle.dump(benchmark_data, f, pickle.HIGHEST_PROTOCOL)


# In[8]:


with open(filename, 'rb') as f:
    bd = pickle.load(f)


# In[9]:


import matplotlib.pyplot as plt


# In[10]:


bd[0].s.shape


# In[11]:


fig, ax = plt.subplots(dpi=150)
fig.patch.set_facecolor('white')

for d in bd:
    ax.plot(d.s, d.phi)

ax.set_xlabel(r'$s$')
ax.set_ylabel(r'$\phi(s)$');


# In[12]:


fig, ax = plt.subplots(dpi=150)
fig.patch.set_facecolor('white')

for d in bd:
    f = d.scattering_amplitude
    ax.plot(np.arccos(d.costheta), np.real(f*np.conj(f)))

ax.set_xlabel(r'$\theta$')
ax.set_ylabel(r'$|f(\theta)|^2$');


# In[ ]:




